
#include <stdio.h>
#include <cstring>
#include <stdlib.h>

extern "C" {
	#include "lua.h"
	#include "lualib.h"
	#include "lauxlib.h"

	extern int luaopen_thelibrary (lua_State* L); // declare the wrapped module
}

#include "thelibrary.h"
/* include the lua runtime so we can actually embed lua scripts */
#include "swig_lua_runtime.h"

/* the lua script as precompiled chunk */
char* LUASCRIPT_PCL = NULL;
size_t LUASCRIPT_PCL_SZ = 0;
size_t LUASCRIPT_PCL_MAXSIZE = 1024;

/* callback to write the data to LUASCRIPT_PCL */
int chunk_writer (lua_State* L, const void* p, size_t sz, void* ud) {
	if (!LUASCRIPT_PCL) {
		LUASCRIPT_PCL = (char *) malloc(LUASCRIPT_PCL_MAXSIZE);
		LUASCRIPT_PCL_SZ = 0;
	}
    memcpy (&LUASCRIPT_PCL[LUASCRIPT_PCL_SZ], p, sz);
	LUASCRIPT_PCL_SZ += sz;
	return 0;
}

/* callback to read the data from LUASCRIPT_PCL */
const char* chunk_reader (lua_State* L, void* data, size_t* sz) {
	*sz = LUASCRIPT_PCL_SZ;
	return (const char*) LUASCRIPT_PCL;
}

int main (int argc, char* argv[]) {
	// locals
	lua_State *L;
	int result;

	if (argc < 2) {
		printf ("%s: <filename.lua>\n", argv[0]);
		return 0;
	}

	printf ("--- precompiling\n\n");

	L = luaL_newstate ();
	luaL_openlibs(L);
	luaopen_thelibrary (L);	// load the wrapped module
	if (luaL_loadfile(L, argv[1]) == 0) {
		// ok, file is loaded, compiled and the precompiled chunk is found at the top of the lua stack
		if (lua_dump(L, &chunk_writer, NULL) == 0) {
			printf ("code precompiled, stored\n");
		}
		else {
			printf ("script error: %s\n", lua_tostring(L, -1));
		};

		// callback into the lua code
		printf("--- callback into lua code\n");
		printf("--- calling: onload\n");
		lua_pcall(L, 0, 0, 0);

		// load callback - function onload ()
		// returns a TheStruct block
		lua_getglobal(L, "onload");
		// do the call, 0 arguments, 1 result
		if (lua_pcall(L, 0, 1, 0) != 0) {
			printf ("onload script error: %s\n", lua_tostring(L, -1));
			return -1;
		}
		// get the result
		if (!lua_isuserdata(L, -1)) {
			printf ("error: function onload() must return a TheStruct instance!\n");
			return -1;
		}
		else {
			printf ("OK: found a userdata on the stack\n");
			// convert userdata to uservalue (this should be the TheStruct* data
			// lua_getuservalue(L, -1);
			int istack = lua_gettop(L);
			printf ("stack top = %d\n", istack);
			printf ("type = %s\n", lua_typename(L, istack));
			printf ("userdata = %d\n", lua_isuserdata(L, istack));
			printf ("lightuserdata = %d\n", lua_islightuserdata(L, istack));
			printf ("table = %d\n", lua_istable(L, istack));
			// get the userdata into the C program
			TheStruct* result;
			swig_type_info* ti=SWIG_TypeQuery(L, "TheStruct");
			SWIG_ConvertPtr(L, istack, (void**)&result, ti, 0);
			// lua_getuservalue(L, istack);
			// read this value into the TheStruct pointer
			lua_pop(L, 1);

			printTheStruct(*result);
		}

		//TheStruct* result;
		//lua_getuservalue(L, -1);
		//printTheStruct(*result);
	}
	lua_close(L);

	printf ("--- running precompiled code\n\n");

	// now create a new state
	L = luaL_newstate ();
	luaL_openlibs(L);
	luaopen_thelibrary (L);	// load the wrapped module
	// now run the precompiled code
	if (LUASCRIPT_PCL) {
		if (lua_load(L, &chunk_reader, NULL, "test", "b") == 0) {
			result = lua_pcall (L, 0, LUA_MULTRET, 0);
			if (result == 0) {
				printf ("executing OK\n");
			}
			else {
				printf ("chunk execute error: %s\n", lua_tostring(L, -1));
			}

			// callback into the lua code
			printf("--- callback into lua code\n\n");
		}
		else {
			printf ("chunck load error: %s\n", lua_tostring(L, -1));
		}
	}
	else {
		printf ("unable to load %s\n", argv[1]);
	}
	lua_close (L);

	return 0;
}
